 /*
  Focuser
  Version: 0.65
  Date:    11/14/11
 */
 
 /*
  PFn                             PWM frequency divider
  MInnnnn                         Move in nnnnn milliseconds 
  MOnnnnn                         Move out nnnnn milliseconds
  ST                              Halt motion
  M?                              Moving? (Y/N)
  V?                              Version? (1.00)
  FInnnnn                         Set full in position nnnnn in milliseconds
  FOnnnnn                         Set full out position nnnnn in milliseconds
  PLnnn                           Set power level in percent nnn
 */

#include <EEPROM.h>

int directionPin1 = 9;    // control motor/direction
int directionPin2 = 10;   // control motor/direction 
char inChar;              // incoming serial byte
char buffer[80];          // command buffer
char buffer2[40];         // short buffer
char buffer3[40];         // short buffer
int bufferPtr = 0;        // position in command buffer

long lastPwmDivisor = 0;  // pulse width modulation speed control
long pwmDivisor = 1;      // pulse width modulation speed control
long moveMS = 0;          // move for this long (milliseconds)
long startTime;
long dir = 0;             // move in this direction (-1 for IN, 0 for STOP, 1 for OUT)
long thisPos = 0;         // current position in ms
long lastPos = 0;         // last position in ms
long fullIn = 100;        // full in position in milliseconds
long fullOut = 100;       // full out position in milliseconds
long powerLevel = 50;     // power level in percent
byte powerLevel1 = 128;   // power level in byte size value

void setup()
{
  // start serial port at 9600 bps:
  Serial.begin(9600);

  // set pins that control the focuser motor
  pinMode(directionPin1, OUTPUT);   // forward
  pinMode(directionPin2, OUTPUT);   // reverse
  digitalWrite(directionPin1,0);
  digitalWrite(directionPin2,0);
  dir=0;

  // recall where we were last time
  lastPos=EEPROM_readLong(0);
  // recall the fi position
  fullIn=EEPROM_readLong(4);
  // recall the fo position
  fullOut=EEPROM_readLong(8);
  // recall the pwm divisor
  pwmDivisor=EEPROM_readLong(12);
  // recall the power level
  powerLevel1=EEPROM_readLong(16);
}

void loop()
{
  int val;
  int div;
  boolean commandError;

  long now;

  // set pwm frequency if it changed
  if (lastPwmDivisor!=pwmDivisor) { 
    switch(pwmDivisor) {
      case 1: div = 1; break;
      case 2: div = 8; break;
      case 3: div = 64; break;
      case 4: div = 256; break;
      case 5: div = 1024; break;
    }
    
    setPwmFrequency(directionPin1, div);
    setPwmFrequency(directionPin2, div);

    lastPwmDivisor=pwmDivisor;
  }

  // keep track of time
  now=(long)millis()-startTime;
  // where we are now in the travel
  if (dir<0) { thisPos=lastPos-now; } else if (dir>0) { thisPos=lastPos+now; }

  if ((dir!=0) && (now<moveMS) && (((thisPos>-fullIn) & (dir<0)) || ((thisPos<fullOut) & (dir>0)))) {
    if (dir==1) {
      // move out
      // no need to adjust pins until we stop
      digitalWrite(directionPin1,0);
      analogWrite(directionPin2,powerLevel1);
      dir=2;
    }
    if (dir==-1) {
      // move in
      digitalWrite(directionPin2,0);
      analogWrite(directionPin1,powerLevel1);
      // no need to adjust pins until we stop
      dir=-2;
    }
  } else {
    // if the timer expires, stop all future motion (the timer may roll over some day)
    if (dir!=0) {
      digitalWrite(directionPin1,0);
      digitalWrite(directionPin2,0);
      dir=0;

      // and record where we are in the travel
      lastPos=thisPos;
      EEPROM_writeLong(0,lastPos);
      
      Serial.println("Stopped");
    }
  }
  
  // if we get a valid byte, read analog ins:
  if (Serial.available() > 0) {
    
    // get incoming byte:
    inChar = Serial.read();
    
    // and build up a command
    buffer[bufferPtr]=toupper(inChar);
    bufferPtr++; 
    buffer[bufferPtr]=(char)0;
    if (bufferPtr>78) { bufferPtr=78; }  // limit maximum command length to avoid overflow

    // process command
    if (inChar==(char)13) {
      buffer[bufferPtr-1]=0;          // erase the cr
      strncpy(buffer2,buffer,2);      // get the two letter command
      strncpy(buffer3,&buffer[2],12); // get the optional numeric tail
      
     // Serial.print(">"); Serial.print(buffer3); Serial.println("<");

      commandError=false;
      
      // set PWM frequency
      if (strcmp(buffer2,"PF") == 0) { 
        if ((dir==0) && (numericTail(&pwmDivisor,1,5))) { EEPROM_writeLong(12,pwmDivisor); Serial.println("Ok"); } else commandError=true;
      } else
      // move in
      if (strcmp(buffer2,"MI") == 0) { 
        if ((dir==0) && (numericTail(&moveMS,0,999999))) { dir=-1; startTime=(long)millis(); Serial.println("Ok"); } else commandError=true;
      } else
      // move out
      if (strcmp(buffer2,"MO") == 0) { 
        if ((dir==0) && (numericTail(&moveMS,0,999999))) { dir=1;  startTime=(long)millis(); Serial.println("Ok"); } else commandError=true;
      } else
      // halt motion
      if (strcmp(buffer2,"ST") == 0) { 
        if (strlen(buffer3)==0) { moveMS=now; Serial.println("Ok"); } else commandError=true;
      } else
      // ask if moving
      if (strcmp(buffer2,"M?") == 0) { 
        if (strlen(buffer3)==0) { if (dir!=0) Serial.println("Y"); else Serial.println("N"); } else commandError=true;
      } else
      // ask for current position
      if (strcmp(buffer2,"P?") == 0) { 
        if (strlen(buffer3)==0) { Serial.println(lastPos); } else commandError=true;
      } else
      // ask for version
      if (strcmp(buffer2,"V?") == 0) { 
        if (strlen(buffer3)==0) { Serial.println("On-Cue Focuser Controller V0.65"); } else commandError=true;
      } else
      // set current position as zero
      if (strcmp(buffer2,"FZ") == 0) { 
        if ((dir==0) && (strlen(buffer3)==0)) { lastPos=0; Serial.println("Ok"); } else commandError=true;
      } else
      // set full in position
      if (strcmp(buffer2,"FI") == 0) { 
        if ((dir==0) && (numericTail(&fullIn,0,999999))) { EEPROM_writeLong(4,fullIn); Serial.println("Ok"); } else commandError=true;
      } else
      // set full out position
      if (strcmp(buffer2,"FO") == 0) { 
        if ((dir==0) && (numericTail(&fullOut,0,999999))) { EEPROM_writeLong(8,fullOut); Serial.println("Ok"); } else commandError=true;
      } else
      // set power level
      if (strcmp(buffer2,"PL") == 0) { 
        if ((dir==0) && (numericTail(&powerLevel,0,100))) { Serial.println("Ok"); powerLevel1=map(powerLevel,0,100,0,255); EEPROM_writeLong(16,powerLevel1); } else commandError=true;
      } else
      // command not recognized
      commandError=true;
      
      if (commandError) {
        moveMS=now; Serial.println("Error");
      }

      bufferPtr=0; buffer[bufferPtr]=0; // clear the last command
    }
  }
}


boolean numericTail(long *val, long low, long high) {
  int l;
  int i;
  boolean isNumeric;
  long temp;

  // get the numeric tail of the command
  l = strlen(buffer3);
  if (l>0) {
    // check to make sure we have a number
    isNumeric=true; for (i=0; i++; i<l) { if ((buffer3[i]<'0') || (buffer3[i]>'9')) { isNumeric=false; } }

    if (isNumeric) {
      temp=atol(buffer3);
      if ((temp>=low) and (temp<=high)) { *val=temp; return true; } else return false;
    } else return false;
    
  } else return false;
}

void setPwmFrequency(int pin, int divisor) {
  byte mode;
  if(pin == 5 || pin == 6 || pin == 9 || pin == 10) {
    switch(divisor) {
      case 1: mode = 0x01; break;
      case 8: mode = 0x02; break;
      case 64: mode = 0x03; break;
      case 256: mode = 0x04; break;
      case 1024: mode = 0x05; break;
      default: return;
    }
    if(pin == 5 || pin == 6) {
      TCCR0B = TCCR0B & 0b11111000 | mode;
    } else {
      TCCR1B = TCCR1B & 0b11111000 | mode;
    }
  } else if(pin == 3 || pin == 11) {
    switch(divisor) {
      case 1: mode = 0x01; break;
      case 8: mode = 0x02; break;
      case 32: mode = 0x03; break;
      case 64: mode = 0x04; break;
      case 128: mode = 0x05; break;
      case 256: mode = 0x06; break;
      case 1024: mode = 0x7; break;
      default: return;
    }
    TCCR2B = TCCR2B & 0b11111000 | mode;
  }
}

// write long numbers into EEPROM
void EEPROM_writeLong(int i,long l) {
  byte *lPtr;
  lPtr = (byte*)&l;
  EEPROM.write(i+0,*lPtr); lPtr++;
  EEPROM.write(i+1,*lPtr); lPtr++;
  EEPROM.write(i+2,*lPtr); lPtr++;
  EEPROM.write(i+3,*lPtr);
}

// read long numbers from EEPROM
long EEPROM_readLong(int i) {
  long l;
  byte *lPtr;
  lPtr = (byte*)&l;
  *lPtr=EEPROM.read(i+0); lPtr++;
  *lPtr=EEPROM.read(i+1); lPtr++;
  *lPtr=EEPROM.read(i+2); lPtr++;
  *lPtr=EEPROM.read(i+3);  
  return l;
}
